import pandas as pa
import numpy as np
from scipy.sparse import csr_matrix, isspmatrix
from MICTI import Micti
import sys

def MICTI(sparceMatrix,geneNames,cellNames,k=None,cluster_assignment=None, th=0,normalized=True, ensembel=False, organisum="hsapiens"):
    #check sparcity of the matrix
    if(sparceMatrix.shape[0]!=len(cellNames)):
        print("The number of cells and the given cell names does not match")
        sys.exit()
    elif(sparceMatrix.shape[1]!=len(geneNames)):
        print("The number of genes and the given gene names does not match")
        sys.exit()
    else:
        #check cluster assignment if it is provided by the user
        if(cluster_assignment is not None):
            if(len(cluster_assignment)!=len(cellNames)):
                print("the number of cells and cluster assignment does not match")
                sys.exit()
            else:
                labelMatrix=pa.get_dummies(cluster_assignment)
                labelArray=np.argwhere(np.array(labelMatrix))[:,1]
                kk=len(set(labelArray))
                cluster_labels=sorted(list(set(cluster_assignment)))
        else:
            labelArray=None
            cluster_labels=None
            kk=k
        if(normalized):
            #change to sparce matrix if the data is not in a sparce matrix format        
            if not isspmatrix(sparceMatrix):
                sparceMatrix=csr_matrix(sparceMatrix)
        else:
            if not isspmatrix(sparceMatrix):
                sparceMatrix=normalizeUMIWithscalefactor(sparceMatrix)
                sparceMatrix=csr_matrix(sparceMatrix)
            else:
                sparceMatrix=normalizeUMIWithscalefactor(sparceMatrix.toarray())
                sparceMatrix=csr_matrix(sparceMatrix)
        #creat micti object
        micti_obj=Micti(sparceMatrix,geneNames,cellNames,k=kk,cluster_label=cluster_labels,cluster_assignment=labelArray, th=th, ensembel=ensembel, organisum=organisum)
    return micti_obj




from MICTI import Kmeans
from MICTI import GM 
import matplotlib.pyplot as plt
import numpy as np
import pandas as pa
from scipy.sparse import csr_matrix, isspmatrix
from sklearn.preprocessing import normalize
from sklearn.metrics import pairwise_distances
import mpl_toolkits.mplot3d.axes3d as p3
import pylab as p
from sklearn.decomposition import TruncatedSVD
from sklearn.manifold import TSNE
import sys
import os
import time
from matplotlib.lines import Line2D
from matplotlib.pyplot import cm
from collections import Counter
from gprofiler import gprofiler
import copy
import operator
import scipy
import seaborn as sns
class Micti:
    def __init__(self,data,geneNames,cellNames,k=None,cluster_label=None,cluster_assignment=None, th=0,seed=None, ensembel=False, organisum="hsapiens"):
        self.data=data
        self.k=k
        self.th=th
        self.geneNames=geneNames
        self.cellNames=cellNames
        self.seed=seed
        self.ensembl=ensembel
        self.organsm=organisum
        self.cluster_assignment=cluster_assignment
        self.cluster_label=cluster_label
        self.data_ICF=self.ICF(self.data)

    def get_cluster_assignment(self):
        return self.cluster_assignment
    
    def get_cluster_data(self, cluster_number):
        return self.data[np.in1d(np.array(self.cluster_assignment), cluster_number),:], self.cellNames[np.in1d(np.array(self.cluster_assignment), cluster_number)]
    def gene_symbol_to_ENSEMBLID(self, symbol, organisum="hsapiens"):
        if(organisum=="hsapiens"):
            genes=pa.read_csv("../data/mart_export_stable_genes.txt", sep="\t")
        elif(organisum=="mmusculus"):
            genes=pa.read_csv("../data/mart_export_mouse_stable_gene.txt", sep="\t")
        else:
            print("give organisum")
       
        genes.index=genes["Gene name"] 
        ENSEBMLID=genes.loc[symbol,"Gene stable ID"]
        return ENSEBMLID.dropna().drop_duplicates()
    
    def ENSEMBLID_to_geneSymbol(self, ENSEMBL, organisum="hsapiens"):
        if(organisum=="hsapiens"):
            genes=pa.read_csv("../data/mart_export_stable_genes.txt", sep="\t")
        elif(organisum=="mmusculus"):
            genes=pa.read_csv("../data/mart_export_mouse_stable_gene.txt", sep="\t")
        else:
            print("give organisum")
        genes.index=genes["Gene stable ID"] 
        gene_symbol=genes.loc[ENSEMBL,"Gene name"]
        return gene_symbol.dropna().drop_duplicates()
    
    def ICF(self,data):
        matrixx=pa.DataFrame((data.T.toarray()))
        totalCells=matrixx.shape[1]
        idf=np.log(totalCells/np.array(matrixx[matrixx > self.th].count(axis=1).add(1)))
        icf_matrix=matrixx.T*np.array(idf)
        return np.array(icf_matrix) 
    
    def get_Visualization(self,dim=2,method="PCA"):
        
        if method=="PCA":
                if dim>3:
                        print ("Please give at most three dimentions")
                else:
                        svd = TruncatedSVD(n_components=dim)
                        svd_fit = svd.fit(self.data)
                        svdTransform=svd.fit_transform(self.data)
                        if dim==3:
                                fig=p.figure()
                                ax = p3.Axes3D(fig)
                                ax.scatter(svdTransform[:,0], svdTransform[:,1], svdTransform[:,2], c=self.cluster_assignment)
                                ax.set_xlabel("PCA1")
                                ax.set_ylabel("PCA2")
                                ax.set_zlabel("PCA3")
                                fig.add_axes(ax)
                                p.show()
                        elif dim==2:
                                plt.scatter(svdTransform[:,0], svdTransform[:,1], c=self.cluster_assignment)
                                plt.xlabel("PCA1")
                                plt.ylabel("PCA2")
                                plt.suptitle("MICTI with k={0:d}".format(self.k), fontsize=8)
                                plt.legend(bbox_to_anchor=(1.65, 1.65), loc='center', ncol=1)
                                plt.legend(list(self.cluster_assignment))
                                plt.show()
                        else:
                                print ("dimentionality error")
        elif method=="tsne":

                if dim>3:
                        print ("Please give at most three dimentions")
                else:
                        svd = TruncatedSVD(n_components=10)
                        svd_fit = svd.fit(self.data)
                        svdTransformTsne=svd.fit_transform(self.data)
                        X_tsne=TSNE(n_components=dim, random_state=0)
                        x_tsne=X_tsne.fit_transform(svdTransformTsne)
                        if dim==3:
                                fig=p.figure()
                                ax = p3.Axes3D(fig)
                                ax.scatter(x_tsne[:,0], x_tsne[:,1], x_tsne[:,2], c=self.cluster_assignment)
                                ax.set_xlabel("tsne1")
                                ax.set_xlabel("tsne2")
                                ax.set_xlabel("tsne3")
                                fig.add_axes(ax)
                                p.show()
                        elif dim==2:
                                data = pa.DataFrame(columns=['tsne_1','tsne_2','cell type'])
                                data['cell type']=[self.cluster_label[i] for i in list(self.cluster_assignment)]
                                data['tsne_1']=x_tsne[:,0]
                                data['tsne_2']=x_tsne[:,1]
                                
                                facet = sns.lmplot(data=data, x='tsne_1', y='tsne_2', hue='cell type', fit_reg=False, legend=True, legend_out=True)
                             
                                plt.savefig("MICTI_Plot.pdf", format="pdf")
                                plt.show()
                        else:
                                print ("dimetionality error")
        else:
                print ("Please give method==pca or method=tsne")

        
    def get_cluster_data(self, cluster_number):
        return self.data[np.in1d(np.array(self.cluster_assignment), cluster_number),:], self.cellNames[np.in1d(np.array(self.cluster_assignment), cluster_number)]
    
    def get_cluster_ICF_data(self, cluster_number):
        return self.ICF(self.data[np.in1d(np.array(self.cluster_assignment), cluster_number),:])
    
    def get_cluster_CF_data(self,cluster_number):
        return self.CF(self.data[np.in1d(np.array(self.cluster_assignment), cluster_number),:]) 
    
    def get_selected_cluster_marker(self, clusters):
        
        datta=self.data[np.in1d(np.array(self.cluster_assignment), clusters),:]
        index=self.cluster_assignment[np.in1d(np.array(self.cluster_assignment), clusters)]
        dat_common=self.CF(datta)
        dat_identity=self.ICF(datta)
        idd_com=[] 
        idd_j=[]
        for j in clusters:
            datt=dat_identity[np.in1d(np.array(index), [j]),:]
            idxx=np.mean(datt, axis=0)
            idxx=np.array(idxx).reshape(idxx.shape[0],)
            idx = idxx.argsort()[::-1]
            iD=[]
            print('Cluster identifier',j)
            if self.ensembl:
                for i in range(18): # Print each gene along with the feature-encoding weight
                    print('{0:s}:{1:.2e}'.format(list(self.ENSEMBLID_to_geneSymbol([self.geneNames[idx[i]]]))[0], idxx[idx[i]]))
                    iD.append(list(self.ENSEMBLID_to_geneSymbol([self.geneNames[idx[i]]]))[0])
            else:
                for i in range(18): # Print each gene along with the feature-encoding weight
                    print('{0:s}:{1:.2e}'.format(self.geneNames[idx[i]], idxx[idx[i]]))
                    iD.append(self.geneNames[idx[i]])
            idd_j.append(iD)
            
        return datt, idxx

    def get_gene_over_representation(self,topn=100):
        enrichmentTable={}    
        for i in range(self.k):
            top10Genes=[] 
            print('Cluster {0:s} ({1:d} cells)'.format(self.cluster_label[i], int(np.sum(self.cluster_assignment==i))))
            idxx=np.mean(self.data_ICF[self.cluster_assignment==i,:], axis=0)
            idxx=np.array(idxx).reshape(idxx.shape[0],)
            idx = idxx.argsort()[::-1]
            for j in range(topn): 
                top10Genes.append(self.geneNames[idx[j]])
            
            if self.ensembl:
                top10Genes=list(self.ENSEMBLID_to_geneSymbol(top10Genes,organisum=self.organsm))
                print(top10Genes)
            else:
                top10Genes=top10Genes
                print(top10Genes)
            enrichment = gprofiler(top10Genes, organism=self.organsm)
            enrichmentTable[i]=enrichment.sort_values(by=['p.value'])[["term.id","p.value","domain","term.name","intersection"]]
            print('')
        return enrichmentTable
    
    def get_MICTI_standardized_mean_over_var(self, clusters):
        
        datta=self.data_ICF[np.in1d(np.array(self.cluster_assignment), clusters),:]
        ccc=np.array(pa.DataFrame(datta).loc[~(pa.DataFrame(datta)==0).all(axis=0)])
        val=np.mean(ccc, axis=0)/(np.log(np.var(ccc, axis=0)+2))
        z_score=(val-np.mean(val))/np.sqrt(np.var(val))
        return z_score
    
    def calculate_pvalue(self, scores):
        return 2*(1-scipy.special.ndtr(abs(scores)))
    def FDR_BH(self, p):
        """Benjamini-Hochberg p-value correction for multiple hypothesis testing."""
        p = np.asfarray(p)
        by_descend = p.argsort()[::-1]
        by_orig = by_descend.argsort()
        steps = float(len(p)) / np.arange(len(p), 0, -1)
        q = np.minimum(1, np.minimum.accumulate(steps * p[by_descend]))
        return q[by_orig]
    def marker_gene_FDR_p_value(self, clusterNo):
        z_score=self.get_MICTI_standardized_mean_over_var(clusterNo)
        p_val=self.calculate_pvalue(z_score)
        FDR_pvalue=self.FDR_BH(p_val)
        result=pa.DataFrame({"Z_scores":z_score,"p_value":p_val,"fdr":FDR_pvalue}, index=self.geneNames)
        return result.sort_values("fdr")
    
    def get_gene_over_representation_for_topn_genes(self,topn=10):
        enrichmentTable={}    
        for i in range(self.k):
            top10Genes=[] 
            print('Cluster {0:s} ({1:d} cells)'.format(str(self.cluster_label[i]), int(np.sum(self.cluster_assignment==i))))
            idxx=np.array(self.marker_gene_FDR_p_value(i)["fdr"])
            idxx=np.array(idxx).reshape(idxx.shape[0],)
            idx = idxx.argsort()
            for j in range(topn): 
                top10Genes.append(self.geneNames[idx[j]])
            
            if self.ensembl:
                top10Genes=list(self.ENSEMBLID_to_geneSymbol(top10Genes,organisum=self.organsm))
                print(top10Genes)
            else:
                top10Genes=top10Genes
                print(top10Genes)
            enrichment = gprofiler(top10Genes, organism=self.organsm)
            enrichmentTable[i]=enrichment.sort_values(by=['p.value'])[["term.id","p.value","domain","term.name","intersection"]]
            print('')
        return enrichmentTable
    def get_gene_list_over_representation_analysis(self, gene_list):
        enrichment = gprofiler(gene_list, organism=self.organsm)
        enrichmentTable=enrichment.sort_values(by=['p.value'])
        return enrichmentTable
    def get_markers_by_Pvalues_and_Zscore(self,cluster,threshold_pvalue=.01, threshold_z_score=0):
        result=self.marker_gene_FDR_p_value(cluster)
        genenames=result.loc[list(np.array(result["fdr"]<threshold_pvalue) & np.array(result["Z_scores"]>threshold_z_score)),:].sort_values("fdr")
        return genenames
    
    def cluster_cells(self,numberOfCluster, method="kmeans"):
        self.cluster_label=[str(i) for i in range(numberOfCluster)]
        if method=="kmeans":
            kmean=Kmeans(self.data, numberOfCluster, self.geneNames, self.cellNames)
            _, self.cluster_assignment=kmean.kmeans_multiple_runs(100,5)
        elif method=="GM":
            EM_GM=GM(self.data, numberOfCluster, self.geneNames, self.cellNames)
            EM_GMM=EM_GM.EM_for_high_dimension()
            self.cluster_assignment=np.argmax(EM_GMM["resp"], axis=1)

        self.get_Visualization(method="tsne")
            
        return None



import numpy as np
import pandas as pa
import time
from sklearn.metrics import pairwise_distances
from scipy.sparse import csr_matrix

class Kmeans:
    def __init__(self,data,k,geneNames,cellNames,cluster_label=None,seed=None):
        self.data=data
        self.k=k
        self.geneNames=geneNames
        self.cellNames=cellNames
        self.seed=seed
        self.centroids=None
        self.cluster_assignment=None
        self.cluster_label=cluster_label
        self.heterogeneity=0.0
        self.get_initial_centroids()
        self.heterogeneities=None

        
    def getCentroids(self):
        return self.centroids
        
    def getCluster_assignment(self):
        return self.cluster_assignment

    def getHeterogenity(self):
        return self.heterogeneity
    
    def getHetrogenities(self):
        return self.heterogeneities
 
    def get_initial_centroids(self):
        '''Randomly choose k data points as initial centroids'''
        if self.seed is not None: # useful for obtaining consistent results
            np.random.seed(self.seed)
        n = self.data.shape[0] # number of data points
        
        # Pick K indices from range [0, N).
        rand_indices = np.random.randint(0, n, self.k)
    
        # Keep centroids as dense format, as many entries will be nonzero due to averaging.
        # As long as at least one document in a cluster contains a word,
        # it will carry a nonzero weight in the TF-IDF vector of the centroid.
        centroids = self.data[rand_indices,:].toarray()
        self.centroids=centroids
        return centroids
    
    def smart_initialize(self):
        '''Use k-means++ to initialize a good set of centroids'''
        if self.seed is not None: # useful for obtaining consistent results
            np.random.seed(self.seed)
        centroids = np.zeros((self.k, self.data.shape[1]))
        # Randomly choose the first centroid.
        # Since we have no prior knowledge, choose uniformly at random
        idx = np.random.randint(self.data.shape[0])
        centroids[0] = self.data[idx,:].toarray()
        # Compute distances from the first centroid chosen to all the other data points
        squared_distances = pairwise_distances(self.data, centroids[0:1], metric='euclidean').flatten()**2
        for i in range(1, self.k):
            # Choose the next centroid randomly, so that the probability for each data point to be chosen
            # is directly proportional to its squared distance from the nearest centroid.
            # Roughtly speaking, a new centroid should be as far as from ohter centroids as possible.
            idx = np.random.choice(self.data.shape[0], 1, p=squared_distances/sum(squared_distances))
            centroids[i] = self.data[idx,:].toarray()
            # Now compute distances from the centroids to all data points
            squared_distances = np.min(pairwise_distances(self.data, centroids[0:i+1], metric='euclidean')**2,axis=1)
        self.centroids=centroids
        return centroids
        
    def assign_clusters(self):
    
        # Compute distances between each data point and the set of centroids:
        # Fill in the blank (RHS only)
        distances_from_centroids = pairwise_distances(self.data,self.centroids,metric='euclidean')
    
        # Compute cluster assignments for each data point:
        # Fill in the blank (RHS only)
        cluster_assignment = np.apply_along_axis(np.argmin, 1, distances_from_centroids)
        self.cluster_assignment=cluster_assignment
        return cluster_assignment
    
    def revise_centroids(self):
        new_centroids = []
        for i in range(self.k):
            # Select all data points that belong to cluster i. Fill in the blank (RHS only)
            member_data_points = self.data[self.cluster_assignment==i]
            # Compute the mean of the data points. Fill in the blank (RHS only)
            centroid = member_data_points.mean(axis=0)
        
            # Convert numpy.matrix type to numpy.ndarray type
            centroid = centroid.A1
        
            new_centroids.append(centroid)
        new_centroids = np.array(new_centroids)
        self.centroids=new_centroids
        return new_centroids
    
    def kmeans(self, maxiter, record_heterogeneity=None, verbose=False):
        '''This function runs k-means on given data and initial set of centroids.
           maxiter: maximum number of iterations to run.
           record_heterogeneity: (optional) a list, to store the history of heterogeneity as function of iterations
                             if None, do not store the history.
           verbose: if True, print how many data points changed their cluster labels in each iteration'''
        centroids = self.centroids[:]
        prev_cluster_assignment = None
        for itr in range(maxiter):        
            if verbose:
                print(itr)
            # 1. Make cluster assignments using nearest centroids
            # YOUR CODE HERE
            cluster_assignment = self.assign_clusters() 
            # 2. Compute a new centroid for each of the k clusters, averaging all data points assigned to that cluster.
            # YOUR CODE HERE
            centroids = self.revise_centroids()  
            # Check for convergence: if none of the assignments changed, stop
            if prev_cluster_assignment is not None and \
              (prev_cluster_assignment==self.cluster_assignment).all():
                break
            # Print number of new assignments 
            if prev_cluster_assignment is not None:
                num_changed = np.sum(prev_cluster_assignment!=self.cluster_assignment)
                if verbose:
                    print('    {0:5d} elements changed their cluster assignment.'.format(num_changed))   
            # Record heterogeneity convergence metric
            if record_heterogeneity is not None:
                # YOUR CODE HERE
                score = compute_heterogeneity(self.data, self.k, centroids, cluster_assignment)
                record_heterogeneity.append(score)
            prev_cluster_assignment = cluster_assignment[:]
        self.centroids=centroids
        self.cluster_assignment=cluster_assignment
        return centroids, cluster_assignment
    
    def kmeans_multiple_runs(self, maxiter, num_runs, seed_list=None, verbose=False):
        
        heterogeneity = {}
        min_heterogeneity_achieved = float('inf')
        best_seed = None
        final_centroids = None
        final_cluster_assignment = None
    
        for i in range(num_runs):
        
            # Use UTC time if no seeds are provided 
            if seed_list is not None: 
                seed = seed_list[i]
                np.random.seed(seed)
            else: 
                seed = int(time.time())
                np.random.seed(seed)
        
            # Use k-means++ initialization
            
            self.initial_centroids = self.smart_initialize()
        
            # Run k-means
 
            centroids, cluster_assignment = self.kmeans(maxiter, record_heterogeneity=None, verbose=False)
        
            # To save time, compute heterogeneity only once in the end
      
            heterogeneity[seed] = self.compute_heterogeneity()
        
            if verbose:
                print('seed={0:06d}, heterogeneity={1:.5f}'.format(seed, heterogeneity[seed]))
                sys.stdout.flush()
        
            # if current measurement of heterogeneity is lower than previously seen,
            # update the minimum record of heterogeneity.
            if heterogeneity[seed] < min_heterogeneity_achieved:
                min_heterogeneity_achieved = heterogeneity[seed]
                best_seed = seed
                final_centroids = centroids
                final_cluster_assignment = cluster_assignment
        self.centroids=final_centroids
        self.cluster_assignment=final_cluster_assignment
        self.heterogeneities=heterogeneity
        return final_centroids, final_cluster_assignment
    
    def clusterEvaluation(self):
        clustMaxDist={}
        clustMinDist={}
        clustMeanDist={}
        for i in range(self.k):
            binMaxDist=[]
            binMinDist=[]
            binMeanDist=[]
            for j in np.concatenate(np.argwhere(self.cluster_assignment==i)):
                dist=pairwise_distances(self.data[np.concatenate(np.argwhere(self.cluster_assignment==i))], self.data[j], metric='euclidean').flatten()
                dist=dist**2
                binMaxDist.append(np.max(dist))
                binMinDist.append(np.min(dist))
                binMeanDist.append(np.mean(dist))
            clustMaxDist[i]=np.max(binMaxDist)
            clustMinDist[i]=np.min(binMinDist)
            clustMeanDist[i]=np.mean(binMeanDist)
        plt.figure(figsize=(7,4.5))
        plt.plot(clustMaxDist.keys(),clustMaxDist.values(), linewidth=2, label='Maximum distance among clusters')
        plt.plot(clustMaxDist.keys(),clustMinDist.values(), linewidth=2, label='Minimum distance among clusters')
        plt.plot(clustMaxDist.keys(),clustMeanDist.values(), linewidth=2, label='avarage distance among clusters')
        plt.xlabel('Cluster number')
        plt.ylabel('Eculidean distance')
        plt.legend(loc='best', prop={'size':15})
        plt.rcParams.update({'font.size':16})
        plt.tight_layout()
        plt.show()
        return np.sum(clustMeanDist)

    def compute_heterogeneity(self):
    
        heterogeneity = 0.0
        for i in range(self.k):
            # Select all data points that belong to cluster i. Fill in the blank (RHS only)
            member_data_points = self.data[self.cluster_assignment==i, :]
        
            if member_data_points.shape[0] > 0: # check if i-th cluster is non-empty
            # Compute distances from centroid to data points (RHS only)
                distances = pairwise_distances(member_data_points, [self.centroids[i]], metric='euclidean')
                squared_distances = distances**2
                heterogeneity += np.sum(squared_distances)
        self.heterogeneity=heterogeneity
        return heterogeneity
        
    def plot_k_vs_heterogeneity(self, k_values, heterogeneity_values):
        plt.figure(figsize=(7,4))
        plt.plot(k_values, heterogeneity_values, linewidth=4)
        plt.xlabel('K')
        plt.ylabel('Heterogeneity')
        plt.title('K vs. Heterogeneity')
        plt.rcParams.update({'font.size': 16})
        plt.tight_layout()
        plt.show()
        return None
    def get_cluster_data(self, cluster_number):
        return self.data[np.in1d(np.array(self.cluster_assignment), cluster_number),:], self.cellNames[np.in1d(np.array(self.cluster_assignment), cluster_number)]
    
    def select_K(self):
        cluster_centroids={}
        cluster_assignments={}
        hetroginity_score=float('inf')
        delta_k={}
        max_K_value=self.k
        hetro_Per_K={}
        deltaHetro=None
        for i in range(max_K_value):
            self.k=i+1
            print("going for k=", i+1)
            cluster_centroid, cluster_assignment=self.kmeans_multiple_runs(5,100)
            hetro=self.compute_heterogeneity()
            hetro_Per_K[i+1]=hetro
            if hetro<hetroginity_score:
                if hetroginity_score==float('inf'):
                    hetroginity_score=hetro
                    deltaHetro=0
                else:
                    deltaHetro=hetroginity_score-hetro
                    hetroginity_score=hetro
                    cluster_centroids[i+1]=cluster_centroid
                    cluster_assignments[i+1]=cluster_assignment
                    delta_k[i+1]=deltaHetro
        best_k=sum(delta_k.values()[1:]>sum(delta_k.values())/(2*len(delta_k.values())))
        print("best k value:", best_k, delta_k)
        self.centroids=cluster_centroids[best_k]
        self.cluster_assignment=cluster_assignments[best_k]
        self.k=best_k
        self.getVisualization(method="tsne")
        self.plot_k_vs_heterogeneity(hetro_Per_K.keys(), hetro_Per_K.values())
        return self.k 




import numpy as np
import pandas as pa
from sklearn.metrics import pairwise_distances
from scipy.sparse import csr_matrix
from scipy.sparse import spdiags
from scipy import stats
from scipy.stats import multivariate_normal
from copy import deepcopy
from MICTI import Kmeans

class GM:
    def __init__(self, data,k,geneNames,cellNames,initial_kmean_Model=None):
        self.data=data
        self.k=k
        self.geneNames=geneNames
        self.cellNames=cellNames
        self.mean=None
        self.weight=None
        self.covs=None
        self.out=None
        self.initial_kmean_Model=initial_kmean_Model
        self.cluster_assignment=None
        self.setMeans()
        self.setWeights()
        self.creatCovMat()
        self.heterogeneity=None
        return 
    
    def diag(self,array):
        n = len(array)
        return spdiags(array, 0, n, n)

    def logpdf_diagonal_gaussian(self, x, mean, cov):
        '''
        Compute logpdf of a multivariate Gaussian distribution with diagonal covariance at a given point x.
        A multivariate Gaussian distribution with a diagonal covariance is equivalent
        to a collection of independent Gaussian random variables.

        x should be a sparse matrix. The logpdf will be computed for each row of x.
        mean and cov should be given as 1D numpy arrays
        mean[i] : mean of i-th variable
        cov[i] : variance of i-th variable'''

        n = x.shape[0]
        dim = x.shape[1]
        assert(dim == len(mean) and dim == len(cov))

        # multiply each i-th column of x by (1/(2*sigma_i)), where sigma_i is sqrt of variance of i-th variable.
        scaled_x = x.dot(self.diag(1./(2*np.sqrt(cov))) )
        # multiply each i-th entry of mean by (1/(2*sigma_i))
        scaled_mean = mean/(2*np.sqrt(cov))

        # sum of pairwise squared Eulidean distances gives SUM[(x_i - mean_i)^2/(2*sigma_i^2)]
        return -np.sum(np.log(np.sqrt(2*np.pi*cov))) - pairwise_distances(scaled_x, [scaled_mean], 'euclidean').flatten()**2
    def initial_cluster_assignment(self):
        initial_kmean_Model=Kmeans(self.data, self.k, self.geneNames, self.cellNames)
        return initial_kmean_Model

    def creatCovMat(self):
        num_clusters=len(self.means)
        covs = []
        for i in range(self.k):
            member_rows = self.data[self.cluster_assignment==i]
            cov = (member_rows.multiply(member_rows) - 2*member_rows.dot(self.diag(self.means[i]))).sum(axis=0).A1 / member_rows.shape[0] \
            + self.means[i]**2
            cov[cov < 1e-8] = 1e-8
            covs.append(cov)
        self.covs=covs
        return covs
    
    def setMeans(self):
        #np.random.seed(5)
        num_clusters = self.k

        # Use scikit-learn's k-means to simplify workflow
        #kmeans_model = KMeans(n_clusters=num_clusters, n_init=5, max_iter=400, random_state=1, n_jobs=-1) # uncomment to use parallelism -- may break on your installation
        if self.initial_kmean_Model == None:
            kmeans_model = self.initial_cluster_assignment()
            centroids, cluster_assignment = kmeans_model.kmeans(10)
        else:
            centroids=self.initial_kmean_Model.centroids
            cluster_assignment=self.initial_kmean_Model.cluster_assignment
            
        means = [centroid for centroid in centroids]
        self.means=means
        self.cluster_assignment=cluster_assignment
        return
    
    def setWeights(self):
        num_docs = self.data.shape[0]
        weights = []
        for i in range(self.k):
        # Compute the number of data points assigned to cluster i:
            num_assigned = self.data[self.cluster_assignment==i,:].shape[0] # YOUR CODE HERE
            w = float(num_assigned) / num_docs
            weights.append(w)
        self.weights=weights
        return 
    
    def log_sum_exp(self,x, axis):
        '''Compute the log of a sum of exponentials'''
        x_max = np.max(x, axis=axis)
        if axis == 1:
            return x_max + np.log( np.sum(np.exp(x-x_max[:,np.newaxis]), axis=1) )
        else:
            return x_max + np.log( np.sum(np.exp(x-x_max), axis=0) )
    
    def EM_for_high_dimension(self,cov_smoothing=1e-5, maxiter=int(1e3), thresh=1e-4, verbose=False):
        # cov_smoothing: specifies the default variance assigned to absent features in a cluster.
        #                If we were to assign zero variances to absent features, we would be overconfient,
        #                as we hastily conclude that those featurese would NEVER appear in the cluster.
        #                We'd like to leave a little bit of possibility for absent features to show up later.
        n = self.data.shape[0]
        dim = self.data.shape[1]
        data=deepcopy(self.data)
        mu = deepcopy(self.means)
        Sigma = deepcopy(self.covs)
        K = len(mu)
        weights = np.array(self.weights)

        ll = None
        ll_trace = []

        for i in range(maxiter):
            # E-step: compute responsibilities
            logresp = np.zeros((n,K))
            for k in range(K):
                logresp[:,k] = np.log(weights[k]) + self.logpdf_diagonal_gaussian(data, mu[k], Sigma[k])
            ll_new = np.sum(self.log_sum_exp(logresp, axis=1))
            if verbose:
                print(ll_new)
            logresp -= np.vstack(self.log_sum_exp(logresp, axis=1))
            resp = np.exp(logresp)
            counts = np.sum(resp, axis=0)

            # M-step: update weights, means, covariances
            weights = counts / np.sum(counts)
            for k in range(K):
                mu[k] = (self.diag(resp[:,k]).dot(data)).sum(axis=0)/counts[k]
                mu[k] = mu[k].A1

                Sigma[k] = self.diag(resp[:,k]).dot( data.multiply(data)-2*data.dot(self.diag(mu[k])) ).sum(axis=0) \
                       + (mu[k]**2)*counts[k]
                Sigma[k] = Sigma[k].A1 / counts[k] + cov_smoothing*np.ones(dim)

            # check for convergence in log-likelihood
            ll_trace.append(ll_new)
            if ll is not None and (ll_new-ll) < thresh and ll_new > -np.inf:
                ll = ll_new
                break
            else:
                ll = ll_new

        out = {'weights':weights,'means':mu,'covs':Sigma,'loglik':ll_trace,'resp':resp}
        self.out=out
        return out
    
    def getResult(self):
        return self.out

    def compute_heterogeneity(self):
        heterogeneity = 0.0
        for i in range(self.k):
            # Select all data points that belong to cluster i. Fill in the blank (RHS only)
            member_data_points = self.data[self.cluster_assignment==i, :]

            if member_data_points.shape[0] > 0: # check if i-th cluster is non-empty
            # Compute distances from centroid to data points (RHS only)
                distances = pairwise_distances(member_data_points, [self.getResult["means"][i]], metric='euclidean')
                squared_distances = distances**2
                heterogeneity += np.sum(squared_distances)
        self.heterogeneity=heterogeneity
        return heterogeneity

    
    def getVisualization(self,dim=2,method="PCA"):

        if method=="PCA":
            if dim>3:
                print("Please give at most three dimentions")
            else:
                svd = TruncatedSVD(n_components=dim)
                svd_fit = svd.fit(self.data)
                svdTransform=svd.fit_transform(self.data)
                if dim==3:
                    fig=p.figure()
                    ax = p3.Axes3D(fig)
                    ax.scatter(svdTransform[:,0], svdTransform[:,1], svdTransform[:,2], c=self.cluster_assignment)
                    ax.set_xlabel("PCA1")
                    ax.set_ylabel("PCA2")
                    ax.set_zlabel("PCA3")
                    fig.add_axes(ax)
                    plt.show()
                elif dim==2:
                    plt.scatter(svdTransform[:,0], svdTransform[:,1], c=self.cluster_assignment)
                    plt.xlabel("PCA1")
                    plt.ylabel("PCA2")
                    plt.suptitle("Gaussian mixture with k={0:d}".format(self.k), fontsize=8)
                    plt.show()
                else:
                    print("dimentionality error")
        elif method=="tsne":
                if dim>3:
                    print("Please give at most three dimentions")
                else:
                    svd = TruncatedSVD(n_components=50)
                    svd_fit = svd.fit(self.data)
                    svdTransformTsne=svd.fit_transform(self.data)
                    X_tsne=TSNE(n_components=dim, random_state=0)
                    x_tsne=X_tsne.fit_transform(svdTransformTsne)
                    if dim==3:
                        fig=p.figure()
                        ax = p3.Axes3D(fig)
                        ax.scatter(x_tsne[:,0], x_tsne[:,1], x_tsne[:,2], c=self.cluster_assignment)
                        ax.set_xlabel("tsne1")
                        ax.set_xlabel("tsne2")
                        ax.set_xlabel("tsne3")
                        fig.add_axes(ax)
                        p.show()
                    elif dim==2:
                        plt.scatter(x_tsne[:,0], x_tsne[:,1], c=self.cluster_assignment)
                        plt.xlabel("tsne1")
                        plt.ylabel("tsne2")
                        plt.suptitle("Gaussian mixture with k={0:d}".format(self.k), fontsize=8)
                        plt.savefig("GM_Plot.png", format="png")
                        plt.show()
                    else:
                        print("dimetionality error")
        else:
            print("Please give method==pca or method=tsne")
                
        return None
    
    def visualize_EM_clusters(self):
        print('')
        print('==========================================================')

        num_clusters = len(self.out["means"])
        for c in range(self.k):
            print('Cluster {0:d}: Largest mean & variance in cluster '.format(c) + "   ")
            print('\n{0: <12}{1: <12}{2: <12}'.format('Gene', 'Mean', 'Variance'))
        
            # The k'th element of sorted_word_ids should be the index of the word 
            # that has the k'th-largest value in the cluster mean. Hint: Use np.argsort().
            sorted_word_ids = np.argsort(-self.out["means"][c])  # YOUR CODE HERE
            print(np.sum(self.cluster_assignment==c))
            for i in sorted_word_ids[:5]:
                print('{0: <12}{1:<10.2e}{2:10.2e}'.format(self.geneNames[i], 
                                                       self.out["means"][c][i],
                                                       self.out["covs"][c][i]))
            print('\n==========================================================')
    
    



import numpy as np 

def normalizeUMIWithscalefactor(data, scale_factor=10e6):
    cellNormalizedData=np.log(1+(data/np.sum(data,axis=1)[0]))*scale_factor
    return(cellNormalizedData)

def getTPM(rowCountData, index_column=None, ensembol_gene=False, withGene_version=False):

    if index_column is not None:
        rowCountData.index=rowCountData[index_column]
        rowCountData=rowCountData.drop([index_column], axis=1)
    #read gene annotation from file
    geneNames=pa.read_csv("data/mart_export_stable_genes_human.txt", sep="\t")
    geneNames["len_inkb"]=(geneNames["Gene end (bp)"]-geneNames["Gene start (bp)"])/1000
    if ensembol_gene & withGene_version:
        geneNames.index=geneNames["Gene stable ID version"]
    elif ensembol_gene:
        geneNames.index=geneNames["Gene stable ID"]
    else:
        geneNames.index=geneNames["Gene name"]
    #get gene length information
    gene_length=geneNames.loc[list(rowCountData.index),:].dropna()["len_inkb"]
    count_data=rowCountData.loc[gene_length.index,:]

    count_data_RKB=count_data.div(list(gene_length), axis=0)
    count_data_TPM=count_data_RKB.div(count_data_RKB.sum(axis=0), axis=1)*10e6

    return count_data_TPM




















	





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="MICTI",
    version="0.1.1",
    author="Nigatu Ayele",
    author_email="naodm2006awet4@gmail.com",
    description="Feature extraction approach in single-cell gene expression profiling for cell-type marker identification.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/insilicolife/micti",
    packages=setuptools.find_packages(), 
    install_requires=['numpy', 'pandas','scipy','scikit-learn', 'matplotlib', 'gprofiler', 'seaborn'],
    classifiers=(
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    )
    
)
